import random
from typing import Optional, Iterator


def backoff(start: float, stop: float, count: int, factor: float = 2.0, jitter: Optional[float] = 1.0) -> Iterator[float]:
    start = float(start)
    stop = float(stop)
    factor = float(factor)
    if jitter:
        jitter = float(jitter)

    if start < 0:
        start = 0.0
    if stop < start:
        stop = start

    current = start
    for i in range(count):
        yield current - (current * jitter * random.random()) if jitter else current

        if current == 0:
            current = 1
        elif current < stop:
            current *= factor

        if current > stop:
            current = stop



from typing import Optional, List, Callable

from ._event import Event
from ._logger import Logger, LoggerPrefixSafeDecorator
from ._options import Options
from ._plugin import Plugin, PluginSafeDecorator
from ._plugin_options import PluginLoadOptions
from ._properties import Properties
from ._validation_response import ValidationResponse

LOG_PREFIX = '[itly-core] '


class Itly:
    def __init__(self) -> None:
        self._options: Optional[Options] = None
        self._plugins: List[Plugin] = []
        self._logger: Logger = Logger.NONE
        self._is_shutdown: bool = False
        self._context: Optional[Event] = None

    def load(self, options: Options) -> None:
        if self._options is not None:
            raise Exception('Itly is already initialized. itly.load() should only be called once.')

        self._options = options
        self._logger = LoggerPrefixSafeDecorator(self._options.logger, LOG_PREFIX)

        if self._options.disabled:
            self._logger.info('disabled = True')
            return

        self._logger.info('load()')

        if options.context is not None:
            self._context = Event(
                name='context',
                properties=options.context,
            )

        for plugin in self._options.plugins:
            plugin_logger = LoggerPrefixSafeDecorator(self._options.logger, f'[plugin-{plugin.id()}] ')
            plugin = PluginSafeDecorator(plugin, plugin_logger)
            self._plugins.append(plugin)
            plugin_options = PluginLoadOptions(environment=self._options.environment, logger=plugin_logger)
            plugin.load(plugin_options)

    def alias(self, user_id: str, previous_id: str) -> None:
        if self._disabled():
            return

        self._logger.info(f'alias(user_id={user_id}, previous_id={previous_id})')
        self._run_on_all_plugins(lambda plugin: plugin.alias(user_id=user_id, previous_id=previous_id))
        self._run_on_all_plugins(lambda plugin: plugin.post_alias(user_id=user_id, previous_id=previous_id))

    def identify(self, user_id: str, identify_properties: Optional[Properties] = None) -> None:
        if self._disabled():
            return

        identify_event = Event('identify', identify_properties)
        self._logger.info(f'identify(user_id={user_id}, properties={identify_event.properties})')
        self._validate_and_run_on_all_plugins(
            identify_event,
            False,
            lambda plugin, event: plugin.identify(user_id, event.properties),
            lambda plugin, event, validation_results: plugin.post_identify(user_id, event.properties, validation_results),
        )

    def group(self, user_id: str, group_id: str, group_properties: Optional[Properties] = None) -> None:
        if self._disabled():
            return

        group_event = Event('group', group_properties)
        self._logger.info(f'group(user_id={user_id}, group_id={group_id}, properties={group_event.properties})')
        self._validate_and_run_on_all_plugins(
            group_event,
            False,
            lambda plugin, event: plugin.group(user_id, group_id, event.properties),
            lambda plugin, event, validation_results: plugin.post_group(user_id, group_id, event.properties, validation_results),
        )

    def page(self, user_id: str, category: Optional[str], name: Optional[str], page_properties: Optional[Properties] = None) -> None:
        if self._disabled():
            return

        page_event = Event('page', page_properties)
        self._logger.info(f'page(user_id={user_id}, category={category}, name={name}, properties={page_event.properties})')
        self._validate_and_run_on_all_plugins(
            page_event,
            False,
            lambda plugin, event: plugin.page(user_id, category, name, event.properties),
            lambda plugin, event, validation_results: plugin.post_page(user_id, category, name, event.properties, validation_results),
        )

    def track(self, user_id: str, event: Event) -> None:
        if self._disabled():
            return

        self._logger.info(f'track(user_id={user_id}, event={event.name}, properties={event.properties})')
        self._validate_and_run_on_all_plugins(
            event,
            True,
            lambda plugin, ev: plugin.track(user_id, ev),
            lambda plugin, ev, validation_results: plugin.post_track(user_id, ev, validation_results),
        )

    def flush(self) -> None:
        if self._disabled():
            return

        self._logger.info('flush()')
        self._run_on_all_plugins(lambda plugin: plugin.flush())

    def shutdown(self) -> None:
        if self._disabled():
            return

        self._logger.info('shutdown()')
        self._is_shutdown = True
        self._run_on_all_plugins(lambda plugin: plugin.shutdown())

    def _validate(self, event: Event) -> List[ValidationResponse]:
        validation_results: List[ValidationResponse] = []

        assert self._options is not None
        if not self._options.validation.disabled:
            for plugin in self._plugins:
                validation_result = plugin.validate(event)
                # Only add invalid validation responses
                if validation_result is not None and not validation_result.valid:
                    validation_results.append(validation_result)

        return validation_results

    def _validate_and_run_on_all_plugins(self,
                                         event: Event,
                                         include_context: bool,
                                         action: Callable[[Plugin, Event], None],
                                         post_action: Callable[[Plugin, Event, List[ValidationResponse]], None]) -> None:
        context_failed_validation_responses = self._validate(self._context) if include_context and self._context is not None else []
        is_context_valid = len(context_failed_validation_responses) == 0

        event_failed_validation_responses = self._validate(event)
        is_event_valid = len(event_failed_validation_responses) == 0

        combined_event: Event = event
        if include_context and self._context is not None:
            combined_event = Event(
                name=event.name,
                properties=Properties.concat([self._context.properties, event.properties]),
                id_=event.id,
                version=event.version,
            )

        assert self._options is not None
        if (is_context_valid and is_event_valid) or self._options.validation.track_invalid:
            self._run_on_all_plugins(lambda plugin: action(plugin, combined_event))

        combined_failed_validation_responses = context_failed_validation_responses + event_failed_validation_responses
        self._run_on_all_plugins(lambda plugin: post_action(plugin, combined_event, combined_failed_validation_responses))

        if (not is_context_valid or not is_event_valid) and self._options.validation.error_on_invalid:
            raise ValueError(combined_failed_validation_responses[0].message)

    def _run_on_all_plugins(self, action: Callable[[Plugin], None]) -> None:
        for plugin in self._plugins:
            action(plugin)

    def _disabled(self) -> bool:
        if self._is_shutdown:
            raise Exception('Itly is shutdown. No more requests are possible.')
        if self._options is None:
            raise Exception('Itly is not yet initialized. Have you called `itly.load()` on app start?')

        return self._options.disabled



from abc import ABC, abstractmethod
import sys


class Logger(ABC):
    STD_OUT_AND_ERR: "Logger" = None  # type: ignore
    NONE: "Logger" = None  # type: ignore

    @abstractmethod
    def debug(self, message: str) -> None:
        pass

    @abstractmethod
    def info(self, message: str) -> None:
        pass

    @abstractmethod
    def warn(self, message: str) -> None:
        pass

    @abstractmethod
    def error(self, message: str) -> None:
        pass


class StdOutAndErrLogger(Logger):
    def debug(self, message: str) -> None:
        print(message)

    def info(self, message: str) -> None:
        print(message)

    def warn(self, message: str) -> None:
        print(message)

    def error(self, message: str) -> None:
        print(message, file=sys.stderr)


class NoneLogger(Logger):
    def debug(self, message: str) -> None:
        pass

    def info(self, message: str) -> None:
        pass

    def warn(self, message: str) -> None:
        pass

    def error(self, message: str) -> None:
        pass


Logger.STD_OUT_AND_ERR = StdOutAndErrLogger()
Logger.NONE = NoneLogger()


class LoggerPrefixSafeDecorator(Logger):
    def __init__(self, logger: Logger, prefix: str, fallback_logger: Logger = Logger.STD_OUT_AND_ERR):
        self._logger = logger
        self._prefix = prefix
        self._fallback_logger = fallback_logger

    def debug(self, message: str) -> None:
        try:
            self._logger.debug(self._prefix + message)
        except Exception as e:
            self._fallback_logger.debug(self._prefix + message)
            self._fallback_logger.error(self._prefix + f'Error in logger.debug(). {e}')

    def info(self, message: str) -> None:
        try:
            self._logger.info(self._prefix + message)
        except Exception as e:
            self._fallback_logger.info(self._prefix + message)
            self._fallback_logger.error(self._prefix + f'Error in logger.info(). {e}')

    def warn(self, message: str) -> None:
        try:
            self._logger.warn(self._prefix + message)
        except Exception as e:
            self._fallback_logger.warn(self._prefix + message)
            self._fallback_logger.error(self._prefix + f'Error in logger.warn(). {e}')

    def error(self, message: str) -> None:
        try:
            self._logger.error(self._prefix + message)
        except Exception as e:
            self._fallback_logger.error(self._prefix + message)
            self._fallback_logger.error(self._prefix + f'Error in logger.error(). {e}')



import enum
import json
from typing import Any, Dict, List, Optional


class Properties:
    def __init__(self, **kwargs: Any):
        self._properties = {key: value for (key, value) in kwargs.items() if value is not None}

    def to_json(self) -> Dict[str, Any]:
        return {
            key: value.value if isinstance(value, enum.Enum) else value
            for (key, value)
            in self._properties.items()
        }

    def __str__(self) -> str:
        return json.dumps(self.to_json())

    def __len__(self) -> int:
        return len(self._properties)

    def __contains__(self, item: str) -> bool:
        return item in self._properties

    @staticmethod
    def concat(properties: List[Optional["Properties"]]) -> Optional["Properties"]:
        if all(p is None for p in properties):
            return None

        result: Dict[str, Any] = {}
        for p in properties:
            if p is not None:
                result.update(p._properties)
        return Properties(**result)



from typing import Optional, List

from .internal import Options as ItlyOptions
from ._environment import Environment
from ._logger import Logger
from ._plugin import Plugin
from ._properties import Properties
from ._validation_options import ValidationOptions


class Options(ItlyOptions):
    def __init__(self,
                 context: Optional[Properties] = None,
                 environment: Environment = Environment.DEVELOPMENT,
                 disabled: bool = False,
                 plugins: Optional[List[Plugin]] = None,
                 validation: Optional[ValidationOptions] = None,
                 logger: Logger = Logger.NONE
                 ):
        super().__init__(
            environment=environment,
            disabled=disabled,
            plugins=plugins,
            validation=validation,
            logger=logger,
        )
        self._context: Optional[Properties] = context

    @property
    def context(self) -> Optional[Properties]:
        return self._context



import queue
from datetime import datetime, timedelta
from threading import Thread, Event
from typing import Optional, Callable, List, Tuple, NamedTuple, Any


class AsyncConsumerMessage(NamedTuple):
    message_type: str
    data: Any


class AsyncConsumer(Thread):
    @staticmethod
    def create_queue() -> queue.Queue:
        return queue.Queue(maxsize=10000)

    def __init__(self, message_queue: queue.Queue, do_upload: Callable[[List[AsyncConsumerMessage]], None], flush_queue_size: int, flush_interval: timedelta) -> None:
        """Create a consumer thread."""
        Thread.__init__(self)
        # Make consumer a daemon thread so that it doesn't block program exit
        self._daemon = True
        self._do_upload = do_upload
        self._upload_size = flush_queue_size
        self._flush_interval = flush_interval
        self._queue = message_queue
        self._pending_message: Optional[AsyncConsumerMessage] = None
        self._running = True

    def run(self) -> None:
        while self._running:
            self.upload()

    def pause(self) -> None:
        self._running = False

    def upload(self) -> None:
        batch, event = self.next()
        if len(batch) == 0:
            if event is not None:
                event.set()
            return

        try:
            self._do_upload(batch)
        except Exception:
            pass
        finally:
            if event is not None:
                event.set()
            # mark items as acknowledged from queue
            for _ in batch:
                self._queue.task_done()

    def next(self) -> Tuple[List[AsyncConsumerMessage], Optional[Event]]:
        start = datetime.now()
        items: List[AsyncConsumerMessage] = [self._pending_message] if self._pending_message is not None else []
        self._pending_message = None

        now = datetime.now()
        while len(items) < self._upload_size and now - start < self._flush_interval:
            try:
                timeout = (self._flush_interval - (now - start)).total_seconds()
                item = self._queue.get(block=True, timeout=timeout)
                if isinstance(item.data, Event):
                    return items, item.data
                if len(items) > 0 and items[-1].message_type != item.message_type:
                    self._pending_message = item
                    break
                items.append(item)
                now = datetime.now()
            except queue.Empty:
                break

        return items, None

    def flush(self) -> None:
        event = Event()
        self._queue.put(AsyncConsumerMessage(message_type='flush', data=event))
        event.wait()

    def shutdown(self) -> None:
        self.pause()
        try:
            self.join()
        except RuntimeError:
            # consumer thread has not started
            pass



from typing import Optional, List

from .._environment import Environment
from .._logger import Logger
from .._plugin import Plugin
from .._validation_options import ValidationOptions

class Options:
    def __init__(self,
                 environment: Environment = Environment.DEVELOPMENT,
                 disabled: bool = False,
                 plugins: Optional[List[Plugin]] = None,
                 validation: Optional[ValidationOptions] = None,
                 logger: Logger = Logger.NONE
                 ):
        self._environment: Environment = environment
        self._disabled: bool = disabled
        self._plugins: List[Plugin] = plugins if plugins is not None else []
        self._validation = validation if validation is not None else ValidationOptions(
            track_invalid=environment == Environment.PRODUCTION,
            error_on_invalid=environment != Environment.PRODUCTION,
        )
        self._logger: Logger = logger

    @property
    def environment(self) -> Environment:
        return self._environment

    @property
    def disabled(self) -> bool:
        return self._disabled

    @property
    def plugins(self) -> List[Plugin]:
        return self._plugins

    @property
    def validation(self) -> ValidationOptions:
        return self._validation

    @property
    def logger(self) -> Logger:
        return self._logger



from abc import ABC, abstractmethod
from typing import Optional, List

from ._event import Event
from ._logger import Logger
from ._plugin_options import PluginLoadOptions
from ._properties import Properties
from ._validation_response import ValidationResponse


class Plugin(ABC):
    # Plugin methods

    @abstractmethod
    def id(self) -> str:
        pass

    def load(self, options: PluginLoadOptions) -> None:
        pass

    # Validation methods

    def validate(self, event: Event) -> Optional[ValidationResponse]:
        return None

    # Tracking methods

    def alias(self, user_id: str, previous_id: str) -> None:
        pass

    def post_alias(self, user_id: str, previous_id: str) -> None:
        pass

    def identify(self, user_id: str, properties: Optional[Properties]) -> None:
        pass

    def post_identify(self, user_id: str, properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        pass

    def group(self, user_id: str, group_id: str, properties: Optional[Properties]) -> None:
        pass

    def post_group(self, user_id: str, group_id: str, properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        pass

    def page(self, user_id: str, category: Optional[str], name: Optional[str], properties: Optional[Properties]) -> None:
        pass

    def post_page(self, user_id: str, category: Optional[str], name: Optional[str], properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        pass

    def track(self, user_id: str, event: Event) -> None:
        pass

    def post_track(self, user_id: str, event: Event, validation_results: List[ValidationResponse]) -> None:
        pass

    def flush(self) -> None:
        pass

    def shutdown(self) -> None:
        pass

    # Helper methods

    def _create_valid_response(self) -> ValidationResponse:
        return ValidationResponse(valid=True, plugin_id=self.id(), message='')

    def _create_invalid_response(self, message: str) -> ValidationResponse:
        return ValidationResponse(valid=False, plugin_id=self.id(), message=message)


class PluginSafeDecorator(Plugin):
    def __init__(self, plugin: Plugin, logger: Logger) -> None:
        self._plugin = plugin
        self._logger = logger

    # Plugin methods

    def id(self) -> str:
        return self._plugin.id()

    def load(self, options: PluginLoadOptions) -> None:
        self._logger.info('load()')
        try:
            self._plugin.load(options)
        except Exception as e:
            self._logger.error(f'Error in load(). {e}')

    # Validation methods

    def validate(self, event: Event) -> Optional[ValidationResponse]:
        if self._plugin.__class__.validate != Plugin.validate:
            self._logger.info(f'validate(event={event.name}, properties={event.properties})')
        try:
            return self._plugin.validate(event)
        except Exception as e:
            self._logger.error(f'Error in validate(). {e}')
            return self._create_invalid_response(message=str(e))

    # Tracking methods

    def alias(self, user_id: str, previous_id: str) -> None:
        if self._plugin.__class__.alias != Plugin.alias:
            self._logger.info(f'alias(user_id={user_id}, previous_id={previous_id})')
        try:
            self._plugin.alias(user_id, previous_id)
        except Exception as e:
            self._logger.error(f'Error in alias(). {e}')

    def post_alias(self, user_id: str, previous_id: str) -> None:
        if self._plugin.__class__.post_alias != Plugin.post_alias:
            self._logger.info(f'post_alias(user_id={user_id}, previous_id={previous_id})')
        try:
            self._plugin.post_alias(user_id, previous_id)
        except Exception as e:
            self._logger.error(f'Error in post_alias(). {e}')

    def identify(self, user_id: str, properties: Optional[Properties]) -> None:
        if self._plugin.__class__.identify != Plugin.identify:
            self._logger.info(f'identify(user_id={user_id}, properties={properties})')
        try:
            self._plugin.identify(user_id, properties)
        except Exception as e:
            self._logger.error(f'Error in identify(). {e}')

    def post_identify(self, user_id: str, properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        if self._plugin.__class__.post_identify != Plugin.post_identify:
            self._logger.info(f'post_identify(user_id={user_id}, properties={properties}, validation_results={validation_results})')
        try:
            self._plugin.post_identify(user_id, properties, validation_results)
        except Exception as e:
            self._logger.error(f'Error in post_identify(). {e}')

    def group(self, user_id: str, group_id: str, properties: Optional[Properties]) -> None:
        if self._plugin.__class__.group != Plugin.group:
            self._logger.info(f'group(user_id={user_id}, group_id={group_id}, properties={properties})')
        try:
            self._plugin.group(user_id, group_id, properties)
        except Exception as e:
            self._logger.error(f'Error in group(). {e}')

    def post_group(self, user_id: str, group_id: str, properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        if self._plugin.__class__.post_group != Plugin.post_group:
            self._logger.info(f'post_group(user_id={user_id}, group_id={group_id}, properties={properties}, validation_results={validation_results})')
        try:
            self._plugin.post_group(user_id, group_id, properties, validation_results)
        except Exception as e:
            self._logger.error(f'Error in post_group(). {e}')

    def page(self, user_id: str, category: Optional[str], name: Optional[str], properties: Optional[Properties]) -> None:
        if self._plugin.__class__.page != Plugin.page:
            self._logger.info(f'page(user_id={user_id}, category={category}, name={name}, properties={properties})')
        try:
            self._plugin.page(user_id, category, name, properties)
        except Exception as e:
            self._logger.error(f'Error in page(). {e}')

    def post_page(self, user_id: str, category: Optional[str], name: Optional[str], properties: Optional[Properties], validation_results: List[ValidationResponse]) -> None:
        if self._plugin.__class__.post_page != Plugin.post_page:
            self._logger.info(f'post_page(user_id={user_id}, category={category}, name={name}, properties={properties}, validation_results={validation_results})')
        try:
            self._plugin.post_page(user_id, category, name, properties, validation_results)
        except Exception as e:
            self._logger.error(f'Error in post_page(). {e}')

    def track(self, user_id: str, event: Event) -> None:
        if self._plugin.__class__.track != Plugin.track:
            self._logger.info(f'track(user_id={user_id}, event={event.name}, properties={event.properties})')
        try:
            self._plugin.track(user_id, event)
        except Exception as e:
            self._logger.error(f'Error in track(). {e}')

    def post_track(self, user_id: str, event: Event, validation_results: List[ValidationResponse]) -> None:
        if self._plugin.__class__.post_track != Plugin.post_track:
            self._logger.info(f'post_track(user_id={user_id}, event={event.name}, properties={event.properties}, validation_results={validation_results})')
        try:
            self._plugin.post_track(user_id, event, validation_results)
        except Exception as e:
            self._logger.error(f'Error in post_track(). {e}')

    def flush(self) -> None:
        if self._plugin.__class__.flush != Plugin.flush:
            self._logger.info('flush()')
        try:
            self._plugin.flush()
        except Exception as e:
            self._logger.error(f'Error in flush(). {e}')

    def shutdown(self) -> None:
        if self._plugin.__class__.shutdown != Plugin.shutdown:
            self._logger.info('shutdown()')
        try:
            self._plugin.shutdown()
        except Exception as e:
            self._logger.error(f'Error in shutdown(). {e}')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['itly', 'itly.sdk', 'itly.sdk.internal']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'itly.sdk',
    'version': '0.1.18',
    'description': 'Iteratively Analytics SDK',
    'long_description': '',
    'author': 'Iteratively',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from ._properties import Properties


class Event:
    def __init__(self, name: str, properties: Optional[Properties] = None, id_: Optional[str] = None, version: Optional[str] = None):
        self._name: str = name
        self._properties: Optional[Properties] = properties
        self._id: Optional[str] = id_
        self._version: Optional[str] = version

    @property
    def name(self) -> str:
        return self._name

    @property
    def properties(self) -> Optional[Properties]:
        return self._properties

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def version(self) -> Optional[str]:
        return self._version



from ._environment import Environment
from ._options import Options
from ._plugin_options import PluginLoadOptions
from ._validation_options import ValidationOptions
from ._validation_response import ValidationResponse
from ._properties import Properties
from ._event import Event
from ._logger import Logger
from ._plugin import Plugin

from ._itly import Itly
